"""Format handler for h5md files."""
import contextlib
import dataclasses
import typing

import h5py
import numpy as np


@dataclasses.dataclass
class GRP:
    """Group names for h5md files.

    Attributes
    ----------
    pbc:
        This group is not a H5MD group that supports time-dependent pbc.
        It can be used to store data from different trajectories in one group,
        that don't share pbc.
    """

    edges: str = "edges"
    boundary: str = "boundary"
    position: str = "position"
    energy: str = "energy"
    species: str = "species"
    forces: str = "forces"
    stress: str = "stress"
    velocity: str = "velocity"
    momentum: str = "momentum"
    pbc: str = "pbc"
    dimension: str = "dimension"

    @staticmethod
    def encode_boundary(value) -> np.ndarray:
        return np.array(
            [
                (
                    "periodic".encode(encoding="utf-8")
                    if x
                    else "none".encode(encoding="utf-8")
                )
                for x in value
            ]
        )

    @staticmethod
    def decode_boundary(value) -> np.ndarray:
        pbc = np.array([x == "periodic".encode(encoding="utf-8") for x in value]).astype(
            bool
        )
        return pbc


PARTICLES_GRP = [
    GRP.position,
    GRP.species,
    GRP.forces,
    GRP.velocity,
    GRP.momentum,
    GRP.pbc,
    GRP.dimension,
    GRP.edges,
    GRP.boundary,
]

OBSERVABLES_GRP = [GRP.energy, GRP.stress]


@dataclasses.dataclass
class StepTimeChunkUnits:
    value: str
    time: str


# We store everything in ASE units
UNITS_PER_GRP = {
    GRP.position: StepTimeChunkUnits(value="Angstrom", time="fs"),
    GRP.velocity: StepTimeChunkUnits(value="Angstrom/fs", time="fs"),
    GRP.forces: StepTimeChunkUnits(value="eV/Angstrom", time="fs"),
    GRP.energy: StepTimeChunkUnits(value="eV", time="fs"),
    GRP.stress: StepTimeChunkUnits(value="GPa", time="fs"),
    GRP.momentum: StepTimeChunkUnits(value="eV/fs", time="fs"),
    GRP.edges: StepTimeChunkUnits(value="Angstrom", time="fs"),
    GRP.boundary: StepTimeChunkUnits(value="Angstrom", time="fs"),
}


@dataclasses.dataclass
class FormatHandler:
    filename: str

    file_handle: typing.Callable = h5py.File

    def __post_init__(self):
        self.particle_key = None
        with self.file as file:
            for particle_key in ["all", "atoms"]:
                if particle_key in file["particles"]:
                    # TODO what if all and atoms appear?
                    self.particle_key = particle_key
        if self.particle_key is None:
            raise ValueError("Could not determine required key '/particles/<...>'")

    @property
    def file(self) -> h5py.File:
        """The 'h5py.File' from filename."""
        return self.file_handle(self.filename)

    @property
    def time_dependent_groups(self) -> typing.List[str]:
        """All time dependent groups.

        References
        ----------
        https://www.nongnu.org/h5md/h5md.html#time-dependent-data

        """
        return list(self.file[f"particles/{self.particle_key}"])

    @property
    def observables_groups(self) -> typing.List[str]:
        """All observables."""
        try:
            return list(self.file[f"observables/{self.particle_key}"])
        except KeyError:
            return []

    def __getattr__(self, item):
        for path in ["particles", "observables"]:
            # we look for the item in particles first and then in observables
            #  TODO: this is mostly legacy and should be removed eventually
            #    because PARTICLES_GRP defines where to look for the item
            group = f"{path}/{self.particle_key}/{item}"
            with contextlib.suppress(KeyError):
                return self.file[group]

        raise AttributeError(f"Could not load group '{group}' from '{self.filename}'")

    @property
    def edges(self):
        group = f"particles/{self.particle_key}/box/{GRP.edges}"
        return self.file[group]

    @property
    def boundary(self):
        group = f"particles/{self.particle_key}/box/{GRP.boundary}"
        return self.file[group]

    @property
    def pbc(self):
        group = f"particles/{self.particle_key}/box/{GRP.pbc}"
        return self.file[group]



from __future__ import annotations

import contextlib

from znh5md.znh5md.base import H5MDBase
from znh5md.znh5md.h5ase import ASEH5MD

with contextlib.suppress(ImportError):
    from znh5md.znh5md.h5dask import DaskH5MD

__all__ = ["H5MDBase", "ASEH5MD", "DaskH5MD"]



from znh5md.io.base import DataReader, DataWriter, ExplicitStepTimeChunk
from znh5md.io.reader import ASEFileReader, AtomsReader, ChemfilesReader

__all__ = [
    "DataReader",
    "DataWriter",
    "ExplicitStepTimeChunk",
    "AtomsReader",
    "ChemfilesReader",
    "ASEFileReader",
]



import contextlib
import dataclasses
import typing

import ase
import h5py
import numpy as np
from ase.calculators.singlepoint import SinglePointCalculator

from znh5md.format import GRP, OBSERVABLES_GRP, PARTICLES_GRP
from znh5md.utils import rm_nan
from znh5md.znh5md.base import H5MDBase


def _gather_value(particles_data, key, idx):
    """Helper to gather the value for a given key and index.

    Returns None if the key is not present in the data.
    """
    try:
        if key in particles_data:
            if key in [GRP.species, GRP.position, GRP.velocity, GRP.forces, GRP.momentum]:
                # use PARTICLES_GRP
                return rm_nan(particles_data[key][idx])
            return particles_data[key][idx]
    except IndexError:
        # a property might not be available at all frames.
        pass
    return None


@dataclasses.dataclass
class ASEH5MD(H5MDBase):
    """ASE interface for H5MD files.

    Attributes
    ----------
    load_all_observables: bool, default=True
        if True, all observables are loaded into 'atoms.calc.results'.
        Otherwise, only the standard ASE observables are loaded.
    """

    load_all_observables: bool = True

    def __getitem__(self, item):
        return self.get_atoms_list(item=item)

    def __getattr__(self, item) -> h5py.Dataset:
        """Get the h5py.Dataset for a given item.

        We only return the value here, because the time and step
        information is not used in the ASE interface.
        """
        # not using dask here was 8x faster on a 32 MB h5 file
        if item == GRP.boundary:
            return getattr(self.format_handler, item)
        return getattr(self.format_handler, item)["value"]

    def _get_particles_dict(self, item) -> dict:
        """Get particles group data."""
        data = {}
        for key in PARTICLES_GRP:
            with contextlib.suppress(AttributeError, KeyError):
                if key == GRP.boundary:
                    data[key] = getattr(self, key)[:]
                else:
                    data[key] = (
                        getattr(self, key)[item] if item else getattr(self, key)[:]
                    )

        if GRP.boundary in data and GRP.pbc not in data:
            data[GRP.pbc] = np.repeat(
                [GRP.decode_boundary(data[GRP.boundary])], len(data[GRP.position]), axis=0
            )

        return data

    def _get_observables_dict(self, item, particles_data: dict) -> dict:
        """Get observables group data."""
        observables = {}

        if self.load_all_observables:
            # TODO this can be removed in future versions
            groups = list(set(self.format_handler.observables_groups + OBSERVABLES_GRP))
        else:
            groups = OBSERVABLES_GRP

        for group_name in groups:
            if group_name not in particles_data:
                with contextlib.suppress(AttributeError, KeyError):
                    # TODO this can be removed in future versions
                    observables[group_name] = (
                        getattr(self, group_name)[item]
                        if item
                        else getattr(self, group_name)[:]
                    )
        return observables

    def get_atoms_list(self, item=None) -> typing.List[ase.Atoms]:
        """Get an 'ase.Atoms' list for all data."""
        single_item = isinstance(item, int)
        if single_item:
            item = [item]

        particles_data = self._get_particles_dict(item=item)
        observables_data = self._get_observables_dict(
            item=item, particles_data=particles_data
        )

        atoms = []

        for idx in range(len(particles_data[GRP.position])):
            obj = ase.Atoms(
                symbols=_gather_value(particles_data, GRP.species, idx),
                positions=_gather_value(particles_data, GRP.position, idx),
                momenta=_gather_value(particles_data, GRP.momentum, idx),
                cell=_gather_value(particles_data, GRP.edges, idx),
                pbc=_gather_value(particles_data, GRP.pbc, idx),
            )
            if GRP.forces in particles_data or len(observables_data):
                obj.calc = SinglePointCalculator(
                    obj, forces=_gather_value(particles_data, GRP.forces, idx)
                )
                for key in observables_data:
                    obj.calc.results[key] = observables_data[key][idx]

            atoms.append(obj)

        return atoms[0] if single_item else atoms



import dataclasses
import functools

import dask.array
import h5py

from znh5md.znh5md.base import H5MDBase


@functools.singledispatch
def get_mask(species: int, data):
    return species == data


@get_mask.register
def _(species: list, data):
    mask = get_mask(species[0], data)
    for x in species[1:]:
        mask += get_mask(x, data)
    return mask


@dataclasses.dataclass
class DaskDataSet:
    value: dask.array
    time: dask.array
    step: dask.array
    species: dask.array

    @classmethod
    def from_file(cls, item: h5py.Dataset, species, value_chunks, time_chunks):
        value = dask.array.from_array(item["value"], chunks=value_chunks)
        time = dask.array.from_array(item["time"], chunks=time_chunks)
        step = dask.array.from_array(item["step"], chunks=time_chunks)
        species = dask.array.from_array(species["value"], chunks=time_chunks)

        if time.shape == ():
            time = dask.array.arange(0, len(value) * time, time)

        if step.shape == ():
            step = dask.array.arange(0, len(value) * step, step)

        return cls(value=value, time=time, step=step, species=species)

    @classmethod
    def from_values(cls, value, time, step, species):
        return cls(value=value, time=time, step=step, species=species)

    def slice_by_species(self, species) -> "DaskDataSet":
        species_flattened = self.species.reshape(-1)
        # species_mask = sum(species_flattened == x for x in species)
        value_flat = self.value.reshape((-1,) + self.value.shape[2:])
        sliced = value_flat[get_mask(species, species_flattened)]
        value = sliced.compute_chunk_sizes().reshape(
            (self.value.shape[0],) + (-1,) + self.value.shape[2:]
        )

        species = (
            species_flattened[get_mask(species, species_flattened)]
            .compute_chunk_sizes()
            .reshape((self.species.shape[0], -1))
        )

        return self.from_values(
            value=value, time=self.time, step=self.step, species=species
        )

    def __getitem__(self, item):
        if isinstance(item, (list, tuple)) and isinstance(item[0], slice):
            return self.from_values(
                value=self.value[item],
                time=self.time[item[0]],
                step=self.step[item[0]],
                species=self.species[item[:2]],
            )
        return self.from_values(
            value=self.value[item],
            time=self.time[item],
            step=self.step[item],
            species=self.species[item],
        )

    def __len__(self) -> int:
        return len(self.value)

    def batch(self, size, axis):
        start = 0
        while start < self.value.shape[axis]:
            if axis == 0:
                yield self[start : start + size]
            elif axis == 1:
                yield self[:, start : start + size]
            elif axis == 2:
                yield self[:, :, start : start + size]
            else:
                raise ValueError(
                    f"axis must be in (0, 1, 2). 'axis={axis}' is currently not"
                    " supported."
                )
            start += size

    @property
    def shape(self):
        return self.value.shape


@dataclasses.dataclass
class DaskDataSet:
    value: dask.array
    time: dask.array
    step: dask.array
    species: dask.array

    @classmethod
    def from_file(cls, item: h5py.Dataset, species, value_chunks, time_chunks):
        value = dask.array.from_array(item["value"], chunks=value_chunks)
        time = dask.array.from_array(item["time"], chunks=time_chunks)
        step = dask.array.from_array(item["step"], chunks=time_chunks)
        species = dask.array.from_array(species["value"], chunks=time_chunks)

        if time.shape == ():
            time = dask.array.arange(0, len(value) * time, time)

        if step.shape == ():
            step = dask.array.arange(0, len(value) * step, step)

        return cls(value=value, time=time, step=step, species=species)

    @classmethod
    def from_values(cls, value, time, step, species):
        return cls(value=value, time=time, step=step, species=species)

    def slice_by_species(self, species) -> "DaskDataSet":
        species_flattened = self.species.reshape(-1)
        # species_mask = sum(species_flattened == x for x in species)
        value_flat = self.value.reshape((-1,) + self.value.shape[2:])
        sliced = value_flat[get_mask(species, species_flattened)]
        value = sliced.compute_chunk_sizes().reshape(
            (self.value.shape[0],) + (-1,) + self.value.shape[2:]
        )

        species = (
            species_flattened[get_mask(species, species_flattened)]
            .compute_chunk_sizes()
            .reshape((self.species.shape[0], -1))
        )

        return self.from_values(
            value=value, time=self.time, step=self.step, species=species
        )

    def __getitem__(self, item):
        if isinstance(item, (list, tuple)) and isinstance(item[0], slice):
            return self.from_values(
                value=self.value[item],
                time=self.time[item[0]],
                step=self.step[item[0]],
                species=self.species[item[:2]],
            )
        return self.from_values(
            value=self.value[item],
            time=self.time[item],
            step=self.step[item],
            species=self.species[item],
        )

    def __len__(self) -> int:
        return len(self.value)

    def batch(self, size, axis):
        start = 0
        while start < self.value.shape[axis]:
            if axis == 0:
                yield self[start : start + size]
            elif axis == 1:
                yield self[:, start : start + size]
            elif axis == 2:
                yield self[:, :, start : start + size]
            else:
                raise ValueError(
                    f"axis must be in (0, 1, 2). 'axis={axis}' is currently not"
                    " supported."
                )
            start += size

    @property
    def shape(self):
        return self.value.shape


@dataclasses.dataclass
class DaskH5MD(H5MDBase):
    """Dask interface for H5MD files.

    Attributes
    ----------
    fixed_species_index: bool, default=False
        if the species indices are fixed in time, a faster way of slicing
        by species is available and can be selected.

    """

    time_chunk_size: int = 10
    species_chunk_size: int = 10
    fixed_species_index: bool = False

    @property
    def position(self) -> DaskDataSet:
        return DaskDataSet.from_file(
            item=self.format_handler.position,
            species=self.format_handler.species,
            value_chunks=(self.time_chunk_size, self.species_chunk_size, 3),
            time_chunks=self.time_chunk_size,
        )

    @property
    def species(self) -> DaskDataSet:
        return DaskDataSet.from_file(
            item=self.format_handler.species,
            species=self.format_handler.species,
            value_chunks=(self.time_chunk_size, self.species_chunk_size),
            time_chunks=self.time_chunk_size,
        )

    def __getattr__(self, item):
        return DaskDataSet.from_file(
            item=getattr(self.format_handler, item),
            species=self.format_handler.species,
            value_chunks="auto",
            time_chunks="auto",
        )



"""ZnH5MD: High Performance Interface for H5MD Trajectories."""
import contextlib
import importlib.metadata

from znh5md import io, utils
from znh5md.format import FormatHandler
from znh5md.znh5md import ASEH5MD

__all__ = ["DaskH5MD", "ASEH5MD", "io", "FormatHandler", "utils"]

with contextlib.suppress(ImportError):
    from znh5md.znh5md.h5dask import DaskH5MD

    __all__.append("DaskH5MD")


__version__ = importlib.metadata.version("znh5md")



import dataclasses
import logging
import typing

import ase.io
import chemfiles
import numpy as np
import tqdm
from ase.calculators.calculator import PropertyNotImplementedError

from znh5md.format import GRP, UNITS_PER_GRP
from znh5md.io.base import DataReader, FixedStepTimeChunk

log = logging.getLogger(__name__)


@dataclasses.dataclass
class AtomsReader(DataReader):
    """Yield ase.Atoms objects from a list of Atoms Objects.

    Parameters
    ----------
    atoms : list[ase.Atoms]
        List of ase.Atoms objects.
    frames_per_chunk : int, optional
        Number of frames to read at once, by default 100
    step : int, optional
        Step size, by default 1
    time : float, optional
        Time step, by default 1
    use_pbc_group : bool, optional
        In addition to the 'boundary' group also
        use the 'pbc' group. This will allow time dependent
        periodic boundary conditions. This is not part of H5MD
        and might cause issues with other software!
    save_atoms_results : bool, optional
        Save 'atoms.calc.results' which can contain custom keys.
    """

    atoms: list[ase.Atoms]
    frames_per_chunk: int = 100  # must be larger than 1
    step: int = 1
    time: float = 1
    use_pbc_group: bool = False
    save_atoms_results: bool = True

    def _get_positions(self, atoms: list[ase.Atoms]) -> np.ndarray:
        data = [x.get_positions() for x in atoms]
        try:
            return np.array(data).astype(float)
        except ValueError:
            return self._fill_with_nan(data).astype(float)

    def _get_energy(self, atoms: list[ase.Atoms]) -> np.ndarray:
        return np.array([x.get_potential_energy() for x in atoms]).astype(float)

    def _get_species(self, atoms: list[ase.Atoms]) -> np.ndarray:
        data = [x.get_atomic_numbers() for x in atoms]
        try:
            return np.array(data).astype(float)
        # NaN is only supported for float, not int
        except ValueError:
            return self._fill_with_nan(data).astype(float)

    def _get_forces(self, atoms: list[ase.Atoms]) -> np.ndarray:
        data = [x.get_forces() for x in atoms]
        try:
            return np.array(data).astype(float)
        except ValueError:
            return self._fill_with_nan(data).astype(float)

    def _get_momenta(self, atoms: list[ase.Atoms]) -> np.ndarray:
        data = [x.arrays["momenta"] for x in atoms]
        try:
            return np.array(data).astype(float)
        except ValueError:
            return self._fill_with_nan(data).astype(float)

    def _get_stress(self, atoms: list[ase.Atoms]) -> np.ndarray:
        return np.array([x.get_stress() for x in atoms]).astype(float)

    def _get_edges(self, atoms: list[ase.Atoms]) -> np.ndarray:
        return np.array([x.get_cell() for x in atoms]).astype(float)

    def _get_pbc(self, atoms: list[ase.Atoms]) -> np.ndarray:
        return np.array([[x.get_pbc()] for x in atoms]).astype(bool)

    def _get_boundary(self, atoms: list[ase.Atoms]) -> np.ndarray:
        data = atoms[0].get_pbc()
        # boundary is constant and should be the same for all atoms
        return GRP.encode_boundary(data)

    def yield_chunks(
        self, group_names: list = None
    ) -> typing.Iterator[typing.Dict[str, FixedStepTimeChunk]]:
        start_index = 0
        stop_index = 0

        pbar = tqdm.tqdm(
            total=len(self.atoms), disable=len(self.atoms) // self.frames_per_chunk < 10
        )

        while stop_index < len(self.atoms):
            stop_index = start_index + self.frames_per_chunk
            data = {}

            functions = {
                GRP.position: self._get_positions,
                GRP.energy: self._get_energy,
                GRP.species: self._get_species,
                GRP.forces: self._get_forces,
                GRP.stress: self._get_stress,
                GRP.edges: self._get_edges,
                GRP.boundary: self._get_boundary,
                GRP.momentum: self._get_momenta,
            }
            if self.use_pbc_group:
                functions[GRP.pbc] = self._get_pbc

            for name in group_names or functions:
                if name not in functions:
                    raise ValueError(f"Value {name} not supported")

                try:
                    value = functions[name](self.atoms[start_index:stop_index])
                    data[name] = FixedStepTimeChunk(
                        value=value,
                        step=self.step,
                        time=self.time,
                        time_units=(
                            UNITS_PER_GRP[name].time if name in UNITS_PER_GRP else None
                        ),
                        value_units=(
                            UNITS_PER_GRP[name].value if name in UNITS_PER_GRP else None
                        ),
                    )
                except (PropertyNotImplementedError, RuntimeError, KeyError) as err:
                    if group_names is not None:
                        # if the property was specifically selected, raise the error
                        raise err
                    else:
                        log.debug(f"Skipping {name} because {err}")

            if self.atoms[0].calc is not None and self.save_atoms_results:
                # we only gather the keys that are present in the first Atoms object.
                # We assume they occur in all the others as well.
                for key in self.atoms[0].calc.results:
                    if key not in functions:
                        value = [x.calc.results[key] for x in self.atoms]
                        try:
                            value = np.array(value).astype(float)
                        except ValueError:
                            value = self._fill_with_nan(value).astype(float)
                        data[key] = FixedStepTimeChunk(
                            value=value,
                            step=self.step,
                            time=self.time,
                            time_units=(
                                UNITS_PER_GRP[GRP.energy].time
                                if GRP.energy in UNITS_PER_GRP
                                else None
                            ),
                            value_units=(
                                UNITS_PER_GRP[GRP.energy].value
                                if GRP.energy in UNITS_PER_GRP
                                else None
                            ),
                        )
            yield data
            start_index = stop_index
            pbar.update(self.frames_per_chunk)

        pbar.close()


@dataclasses.dataclass
class ASEFileReader(DataReader):
    """Use ASE to read files.

    Parameters
    ----------
    filename : str
        Path to the file. Any format supported by ASE is supported.
    frames_per_chunk : int, optional
        Number of frames to read at once
    time : float, optional
        Time step
    step : int, optional
        Step size
    """

    filename: str
    frames_per_chunk: int = 5000
    time: float = 1
    step: int = 1
    use_pbc_group: bool = False

    def yield_chunks(self) -> typing.Iterator[typing.Dict[str, FixedStepTimeChunk]]:
        """Yield chunks using AtomsReader."""
        atoms_list = []
        reader = AtomsReader(
            None,  # we set atoms in the reading loop
            frames_per_chunk=self.frames_per_chunk,
            time=self.time,
            step=self.step,
            use_pbc_group=self.use_pbc_group,
        )

        for atoms in tqdm.tqdm(ase.io.iread(self.filename)):
            atoms_list.append(atoms)
            if len(atoms_list) == self.frames_per_chunk:
                reader.atoms = atoms_list
                yield from reader.yield_chunks()
                atoms_list = []
        if len(atoms_list) > 0:
            reader.atoms = atoms_list
            yield from reader.yield_chunks()


@dataclasses.dataclass
class ChemfilesReader(DataReader):
    filename: str
    format: str = ""
    frames_per_chunk: int = 5000
    step: int = 1
    time: float = 1

    def yield_chunks(self) -> typing.Iterator[typing.Dict[str, FixedStepTimeChunk]]:
        with chemfiles.Trajectory(self.filename, format=self.format) as trajectory:
            positions = []
            species = []
            energy = []
            cell = []
            pbc = []
            for frame in tqdm.tqdm(trajectory, total=trajectory.nsteps):
                positions.append(np.copy(frame.positions))
                species.append(np.copy([atom.atomic_number for atom in frame.atoms]))
                cell.append(np.copy(frame.cell.lengths))

                if "energy" in frame.list_properties():
                    energy.append(np.copy(frame["energy"]).astype(float))
                if "pbc" in frame.list_properties():
                    pbc.append(
                        [True if x == "T" else False for x in frame["pbc"].split()]
                    )

                if len(positions) == self.frames_per_chunk:
                    positions = self._fill_with_nan(positions).astype(float)
                    species = self._fill_with_nan(species).astype(float)

                    data = {
                        GRP.position: FixedStepTimeChunk(
                            value=positions,
                            step=self.step,
                            time=self.time,
                        ),
                        GRP.species: FixedStepTimeChunk(
                            value=species,
                            step=self.step,
                            time=self.time,
                        ),
                        GRP.edges: FixedStepTimeChunk(
                            value=np.stack(cell),
                            step=self.step,
                            time=self.time,
                        ),
                    }
                    if len(energy) > 0:
                        data[GRP.energy] = FixedStepTimeChunk(
                            value=np.stack(energy),
                            step=self.step,
                            time=self.time,
                        )
                    if len(pbc) > 0:
                        data[GRP.pbc] = FixedStepTimeChunk(
                            value=np.stack(pbc),
                            step=self.step,
                            time=self.time,
                        )

                    yield data

                    positions = []
                    species = []
                    energy = []
                    cell = []
                    pbc = []
            if len(positions) > 0:
                positions = self._fill_with_nan(positions).astype(float)
                species = self._fill_with_nan(species).astype(float)

                data = {
                    GRP.position: FixedStepTimeChunk(
                        value=positions,
                        step=self.step,
                        time=self.time,
                    ),
                    GRP.species: FixedStepTimeChunk(
                        value=species,
                        step=self.step,
                        time=self.time,
                    ),
                    GRP.edges: FixedStepTimeChunk(
                        value=np.stack(cell),
                        step=self.step,
                        time=self.time,
                    ),
                }
                if len(energy) > 0:
                    data[GRP.energy] = FixedStepTimeChunk(
                        value=np.stack(energy),
                        step=self.step,
                        time=self.time,
                    )
                if len(pbc) > 0:
                    data[GRP.pbc] = FixedStepTimeChunk(
                        value=np.stack(pbc),
                        step=self.step,
                        time=self.time,
                    )
                yield data



import dataclasses
import os
import pathlib
import typing

from znh5md.format import FormatHandler

PATHLIKE = typing.Union[str, pathlib.Path, os.PathLike]


@dataclasses.dataclass
class H5MDBase:
    filename: PATHLIKE
    format_handler: FormatHandler = FormatHandler

    def __post_init__(self):
        self.format_handler = self.format_handler(self.filename)



import abc
import dataclasses
import logging
import pathlib
import typing

import h5py
import numpy as np

log = logging.getLogger(__name__)

import typing_extensions as te

from znh5md.format import GRP, PARTICLES_GRP


@dataclasses.dataclass
class StepTimeChunk:
    """Abstract class for time-dependent data for a single group.

    Parameters
    ----------
    value : np.ndarray
        The value, to be stored in the value dataset.
    step : np.ndarray
        The step, to be stored in the step dataset.
    time : np.ndarray
        The time, to be stored in the time dataset.

    References
    ----------
    https://h5md.nongnu.org/h5md.html#time-dependent-data
    """

    value: np.ndarray
    step: np.ndarray
    time: np.ndarray

    value_units: str = None
    time_units: str = None

    @property
    def shape(self) -> tuple:
        """The shape of the value array."""
        return tuple(None for _ in range(len(self.value.shape)))

    def __len__(self):
        """Get the number of frames in the chunk."""
        return len(self.value)

    def create_dataset(self, dataset_group: h5py.Group):
        """Create the datasets for the chunk."""
        raise NotImplementedError

    def append_to_dataset(self, dataset_group: h5py.Group):
        """Append the data to the dataset."""
        raise NotImplementedError

    def _resize_dataset_group(self, dataset_group: h5py.Group, fill_value):
        """Resize the dataset_group."""
        old_size = dataset_group["value"].shape[1]
        dataset_group["value"].resize(self.value.shape[1], axis=1)
        dataset_group["value"][:, old_size:] = fill_value

    def _resize_value(self, n_new_particles: int, fill_value):
        """Resize the value array."""
        fill_shape = list(self.value.shape)
        fill_shape[1] = n_new_particles
        self.value = np.concatenate([self.value, np.full(fill_shape, fill_value)], axis=1)

    def resize_by_particle_count(self, dataset_group: h5py.Group, fill_value=np.nan):
        # We also have to reshape value, if the the shape
        #  changed in axis=1 (e.g. number of atoms)
        if len(self.value.shape) > 1:
            if self.value.shape[1] > dataset_group["value"].shape[1]:
                # we resize the group
                # we fill the new values with Nan
                self._resize_dataset_group(dataset_group, fill_value)

            elif self.value.shape[1] < dataset_group["value"].shape[1]:
                # we add Nan to the chunk data, because it is smaller than the group
                n_new_particles = dataset_group["value"].shape[1] - self.value.shape[1]
                self._resize_value(n_new_particles, fill_value)


@dataclasses.dataclass
class ExplicitStepTimeChunk(StepTimeChunk):
    """Same as StepTimeChunk, but with explicit step and time."""

    def create_dataset(self, dataset_group: h5py.Group):
        """Create the datasets for the chunk."""
        value_ds = dataset_group.create_dataset(
            "value", maxshape=self.shape, data=self.value, chunks=True
        )
        time_ds = dataset_group.create_dataset(
            "time", maxshape=(None,), data=self.time, chunks=True
        )
        dataset_group.create_dataset(
            "step", maxshape=(None,), data=self.step, chunks=True
        )

        if self.value_units is not None:
            value_ds.attrs["unit"] = self.value_units
        if self.time_units is not None:
            time_ds.attrs["unit"] = self.time_units

    def append_to_dataset(self, dataset_group: h5py.Group):
        n_current_frames = dataset_group["value"].shape[0]

        self.resize_by_particle_count(dataset_group)
        for key in ("value", "time", "step"):
            dataset_group[key].resize(n_current_frames + len(self), axis=0)
            dataset_group[key][:] = np.concatenate(
                [dataset_group[key][:n_current_frames], self.value]
            )


@dataclasses.dataclass
class FixedStepTimeChunk(StepTimeChunk):
    """Same as StepTimeChunk, but with fixed step and time."""

    step: int
    time: float

    def create_dataset(self, dataset_group: h5py.Group):
        """Create the datasets for the chunk."""
        value_ds = dataset_group.create_dataset(
            "value", maxshape=self.shape, data=self.value, chunks=True
        )
        time_ds = dataset_group.create_dataset(
            "time",
            data=np.arange(len(self.value)) * self.time,
            chunks=True,
            maxshape=(None,),
        )
        dataset_group.create_dataset(
            "step", data=np.arange(len(self.value)), chunks=True, maxshape=(None,)
        )

        if self.value_units is not None:
            value_ds.attrs["unit"] = self.value_units
        if self.time_units is not None:
            time_ds.attrs["unit"] = self.time_units

    def append_to_dataset(self, dataset_group: h5py.Group):
        n_current_frames = dataset_group["value"].shape[0]

        self.resize_by_particle_count(dataset_group)
        dataset_group["value"].resize(n_current_frames + len(self), axis=0)
        dataset_group["value"][:] = np.concatenate(
            [dataset_group["value"][:n_current_frames], self.value]
        )
        # append to time and step as well
        dataset_group["time"].resize(n_current_frames + len(self), axis=0)
        dataset_group["time"][:] = np.concatenate(
            [
                dataset_group["time"][:n_current_frames],
                np.arange(len(self.value)) * self.time + n_current_frames * self.time,
            ]
        )
        dataset_group["step"].resize(n_current_frames + len(self), axis=0)
        dataset_group["step"][:] = np.concatenate(
            [
                dataset_group["step"][:n_current_frames],
                np.arange(len(self.value)) + n_current_frames,
            ]
        )


CHUNK_DICT = typing.Dict[str, ExplicitStepTimeChunk]


class DataReader(abc.ABC):
    """Abstract base class for reading data and yielding chunks."""

    @abc.abstractmethod
    def yield_chunks(
        self, *args, **kwargs
    ) -> typing.Iterator[typing.Dict[str, StepTimeChunk]]:
        """Yield chunks of data.

        This method will yield chunks of data to be written to the HDF5 File.
        It should implement a generator pattern that e.g. reads from files.

        Returns
        -------
        typing.Iterator[typing.Dict[str, ExplicitStepTimeChunk]]
            A dictionary of chunks. The key is the name of the group.
            Each chunk containing the data for one group.
        """
        raise NotImplementedError()

    def _fill_with_nan(self, data: list) -> np.ndarray:
        max_n_particles = max(x.shape[0] for x in data)
        dimensions = data[0].shape[1:]

        result = np.full((len(data), max_n_particles, *dimensions), np.nan)
        for i, x in enumerate(data):
            result[i, : x.shape[0], ...] = x
        return result


@dataclasses.dataclass
class DataWriter:
    filename: str
    particles_path: str = "particles/atoms"
    observables_path: str = "observables/atoms"

    @te.deprecated("DB will be initialized automatically when adding data.")
    def initialize_database_groups(self):
        """Create all groups that are required.

        We create the following groups:
        - particles/atoms
        """
        pass

    def _handle_special_cases_group_names(self, groupname: str) -> str:
        """Update group name in special cases.

        Some groups, especially the box group, are nested differently.
        """
        if groupname in [GRP.boundary, GRP.edges, GRP.pbc]:
            return f"box/{groupname}"

        return groupname

    def create_group(self, db_path, group_name, chunk_data):
        """Create a new group for the given elements.

        This method will create the following datasets for each group in kwargs.
        - <db_path>/<group_name>/value
        - <db_path>/<group_name>/time
        - <db_path>/<group_name>/step

        Parameters
        ----------
        kwargs: dict[str, ExplicitStepTimeChunk]
            The chunk data to write to the database. The key is the name of the group.
        """
        group_name = self._handle_special_cases_group_names(group_name)
        dataset_group = db_path.create_group(group_name)
        chunk_data.create_dataset(dataset_group)

    def add_data_to_group(self, db_path, group_name, chunk_data: StepTimeChunk):
        """Add data to an existing group.

        For each group in kwargs, the following datasets are resized and appended to:
        - <db_path>/<group_name>/value
        - <db_path>/<group_name>/time
        - <db_path>/<group_name>/step

        Parameters
        ----------
        kwargs: dict[str, ExplicitStepTimeChunk]
            The chunk data to write to the database. The key is the name of the group.
            The group must already exist.
        """
        group_name = self._handle_special_cases_group_names(group_name)
        dataset_group = db_path[group_name]
        chunk_data.append_to_dataset(dataset_group)

    def handle_boundary(self, file, chunk_data):
        """Special case for the box boundary.

        This requires a special case, because the boundary is inside
        the box group.
        """
        if GRP.boundary not in file[f"{self.particles_path}/box"]:
            atoms = file[self.particles_path]
            # we create the box group
            atoms.create_dataset(f"box/{GRP.boundary}", data=chunk_data.value)
            # dimension group is required by H5MD
            atoms.create_dataset(f"box/{GRP.dimension}", data=len(chunk_data.value))

            atoms["box"].attrs["boundary"] = ["periodic", "periodic", "periodic"]
            atoms["box"].attrs["dimension"] = 3

    def add_chunk_data(self, **kwargs: CHUNK_DICT) -> None:
        """Write Chunks to the database.

        Create a new group, if it does not exist yet.
        Add to existing groups otherwise.

        Parameters
        ----------
        kwargs: dict[str, ExplicitStepTimeChunk]
            The chunk data to write to the database. The key is the name of the group.
        """
        if not pathlib.Path(self.filename).exists():
            _ = h5py.File(self.filename, "w")  # create the file
        with h5py.File(self.filename, "r+") as file:
            for group_name, chunk_data in kwargs.items():
                if group_name == GRP.boundary:
                    self.handle_boundary(file, chunk_data)
                else:
                    if group_name in PARTICLES_GRP:
                        try:
                            group_path = file[self.particles_path]
                        except KeyError:
                            log.debug(f"creating particle groups {group_name}")
                            group_path = file.create_group(self.particles_path)
                    else:
                        try:
                            group_path = file[self.observables_path]
                        except KeyError:
                            log.debug(f"creating observable groups {group_name}")
                            group_path = file.create_group(self.observables_path)

                    try:
                        self.add_data_to_group(group_path, group_name, chunk_data)
                    except KeyError:
                        log.debug(f"creating particle groups {group_name}")

                        self.create_group(group_path, group_name, chunk_data)

    def add(self, reader: DataReader):
        """Add data from a reader to the HDF5 file."""
        for chunk in reader.yield_chunks():
            self.add_chunk_data(**chunk)
